from minos.common import (
    Config,
    Inject,
)
from minos.networks import (
    BrokerRequest,
    EnrouteDecorator,
    enroute,
)
from minos.saga import (
    SagaManager,
    SagaResponse,
)


class SagaService:
    """Saga Service class"""

    # noinspection PyUnusedLocal
    @Inject()
    def __init__(self, *args, saga_manager: SagaManager, **kwargs):
        self.saga_manager = saga_manager

    @classmethod
    def __get_enroute__(cls, config: Config) -> dict[str, set[EnrouteDecorator]]:
        return {cls.__reply__.__name__: {enroute.broker.command(f"{config.get_name()}Reply")}}

    async def __reply__(self, request: BrokerRequest) -> None:
        response = SagaResponse.from_message(request.raw)
        await self.saga_manager.run(response=response, pause_on_disk=True, raise_on_error=False, return_execution=False)



from __future__ import (
    annotations,
)

import warnings
from typing import (
    Any,
    Iterable,
    Optional,
    Type,
    TypeVar,
    Union,
)

from ..exceptions import (
    AlreadyCommittedException,
    AlreadyOnSagaException,
    EmptySagaException,
    SagaNotCommittedException,
)
from .operations import (
    SagaOperation,
)
from .steps import (
    ConditionalSagaStep,
    LocalSagaStep,
    RemoteSagaStep,
    SagaStep,
)
from .types import (
    LocalCallback,
    RequestCallBack,
)


class Saga:
    """Saga class.

    The purpose of this class is to define a sequence of operations among microservices.
    """

    # noinspection PyUnusedLocal
    def __init__(self, *args, steps: list[SagaStep] = None, committed: bool = False, commit: None = None, **kwargs):
        if steps is None:
            steps = list()

        self.steps = steps
        self.committed = committed

        if commit is not None:
            warnings.warn(f"Commit callback is being deprecated. Use {self.local_step!r} instead", DeprecationWarning)
            self.local_step(commit)
            self.committed = True

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], Saga], **kwargs) -> Saga:
        """Build a new ``Saga`` instance from raw.

        :param raw: The raw representation of the saga.
        :param kwargs: Additional named arguments.
        :return: A new ``Saga`` instance.
        """
        if isinstance(raw, cls):
            return raw

        current = raw | kwargs

        steps = [SagaStep.from_raw(step) for step in current.pop("steps")]

        instance = cls(steps=steps, **current)

        return instance

    def conditional_step(self, step: Optional[ConditionalSagaStep] = None) -> ConditionalSagaStep:
        """Add a new conditional step.

        :param step: The step to be added. If `None` is provided then a new one will be created.
        :return: A ``SagaStep`` instance.
        """
        return self._add_step(ConditionalSagaStep, step)

    def local_step(
        self, step: Optional[Union[LocalCallback, SagaOperation[LocalCallback], LocalSagaStep]] = None, **kwargs
    ) -> LocalSagaStep:
        """Add a new local step.

        :param step: The step to be added. If `None` is provided then a new one will be created.
        :param kwargs: Additional named parameters.
        :return: A ``SagaStep`` instance.
        """
        if step is not None and not isinstance(step, SagaStep) and not isinstance(step, SagaOperation):
            step = SagaOperation(step, **kwargs)
        return self._add_step(LocalSagaStep, step)

    def step(
        self, step: Optional[Union[RequestCallBack, SagaOperation[RequestCallBack], RemoteSagaStep]] = None, **kwargs
    ) -> RemoteSagaStep:
        """Add a new remote step step.

        :param step: The step to be added. If `None` is provided then a new one will be created.
        :param kwargs: Additional named parameters.
        :return: A ``SagaStep`` instance.
        """
        warnings.warn("step() method is deprecated by remote_step() and will be removed soon.", DeprecationWarning)
        return self.remote_step(step, **kwargs)

    def remote_step(
        self, step: Optional[Union[RequestCallBack, SagaOperation[RequestCallBack], RemoteSagaStep]] = None, **kwargs
    ) -> RemoteSagaStep:
        """Add a new remote step step.

        :param step: The step to be added. If `None` is provided then a new one will be created.
        :param kwargs: Additional named parameters.
        :return: A ``SagaStep`` instance.
        """
        if step is not None and not isinstance(step, SagaStep) and not isinstance(step, SagaOperation):
            step = SagaOperation(step, **kwargs)
        return self._add_step(RemoteSagaStep, step)

    def _add_step(self, step_cls: Type[T], step: Optional[Union[SagaOperation, T]]) -> T:
        if self.committed:
            raise AlreadyCommittedException("It is not possible to add more steps to an already committed saga.")

        if step is None:
            step = step_cls(saga=self)
        elif isinstance(step, SagaStep):
            if not isinstance(step, step_cls):
                raise TypeError(f"The given step does not match the requested type: {step_cls} vs {type(step)}")
            if step.saga is not None:
                raise AlreadyOnSagaException()
            step.saga = self
        else:
            step = step_cls(step, saga=self)

        self.steps.append(step)
        return step

    @property
    def raw(self) -> dict[str, Any]:
        """Generate a raw representation of the instance.

        :return: A ``dict`` instance.
        """
        ans = {
            "steps": [step.raw for step in self.steps],
            "committed": self.committed,
        }
        return ans

    def __eq__(self, other: SagaStep) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __iter__(self) -> Iterable:
        yield from (
            self.steps,
            self.committed,
        )

    def commit(self, callback: None = None, **kwargs) -> Saga:
        """Commit the instance to be ready for execution.

        :param callback: Deprecated argument.
        :param kwargs: Additional named arguments.
        :return: A ``Saga`` instance.
        """
        if self.committed:
            raise AlreadyCommittedException("It is not possible to commit a saga multiple times.")

        if callback is not None:
            warnings.warn(f"Commit callback is being deprecated. Use {self.local_step!r} instead", DeprecationWarning)
            self.local_step(callback, **kwargs)

        self.committed = True
        return self

    def validate(self) -> None:
        """Check if the saga is valid.

        :return: This method does not return anything, but raises an exception if the saga is not valid.
        """

        if not len(self.steps):
            raise EmptySagaException()

        for step in self.steps:
            step.validate()

        if not self.committed:
            raise SagaNotCommittedException()


T = TypeVar("T")



from __future__ import (
    annotations,
)

from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Iterable,
    Optional,
    Union,
)

from minos.common import (
    classname,
)

from ...exceptions import (
    EmptySagaStepException,
    MultipleElseThenException,
)
from ..operations import (
    SagaOperation,
)
from .abc import (
    SagaStep,
)

if TYPE_CHECKING:
    from ...context import (
        SagaContext,
    )
    from ..saga import (
        Saga,
    )


class ConditionalSagaStep(SagaStep):
    """Conditional Saga Step class."""

    if_then_alternatives: list[IfThenAlternative]
    else_then_alternative: ElseThenAlternative

    def __init__(
        self,
        if_then: Optional[Union[IfThenAlternative, Iterable[IfThenAlternative]]] = None,
        else_then: Optional[ElseThenAlternative] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if isinstance(if_then, IfThenAlternative):
            if_then = [if_then]

        if if_then is None:
            if_then = list()

        self.if_then_alternatives = list(if_then)
        self.else_then_alternative = else_then

    @classmethod
    def _from_raw(cls, raw: dict[str, Any], **kwargs) -> ConditionalSagaStep:
        current = raw | kwargs

        if_then = current.get("if_then", None)
        if if_then is not None:
            current["if_then"] = [IfThenAlternative.from_raw(alternative) for alternative in if_then]

        else_then = current.get("else_then", None)
        if else_then is not None:
            current["else_then"] = ElseThenAlternative.from_raw(else_then)

        return cls(**current)

    def if_then(
        self, condition: Callable[[SagaContext], Union[bool, Awaitable[bool]]], saga: Saga
    ) -> ConditionalSagaStep:
        """Add a new ``IfThenAlternative`` based on a condition and a saga.

        :param condition: The condition that must be satisfied to execute the alternative.
        :param saga: The saga to be executed if the condition is satisfied.
        :return: This method returns the same instance that is called.
        """

        alternative = IfThenAlternative(condition, saga)
        self.if_then_alternatives.append(alternative)
        return self

    def else_then(self, saga: Saga) -> ConditionalSagaStep:
        """Set the ``ElseThenAlternative`` with the given saga.

        :param saga: The saga to be executed if not any condition is met.
        :return: This method returns the same instance that is called.
        """

        if self.else_then_alternative is not None:
            raise MultipleElseThenException()

        alternative = ElseThenAlternative(saga)
        self.else_then_alternative = alternative
        return self

    def validate(self) -> None:
        """Check if the step is valid.

        :return: This method does not return anything, but raises an exception if the step is not valid.
        """

        if not len(self.if_then_alternatives) and self.else_then_alternative is None:
            raise EmptySagaStepException()

        for alternative in self.if_then_alternatives:
            alternative.validate()

        if self.else_then_alternative is not None:
            self.else_then_alternative.validate()

    @property
    def raw(self) -> dict[str, Any]:
        """Get the raw representation of the step.

        :return: A ``dict`` instance.
        """

        return {
            "cls": classname(type(self)),
            "if_then": [alternative.raw for alternative in self.if_then_alternatives],
            "else_then": None if self.else_then_alternative is None else self.else_then_alternative.raw,
        }

    def __iter__(self) -> Iterable:
        yield from (
            self.if_then_alternatives,
            self.else_then_alternative,
        )


class IfThenAlternative:
    """If Then Alternative class."""

    def __init__(
        self, condition: Union[SagaOperation, Callable[[SagaContext], Union[bool, Awaitable[bool]]]], saga: Saga
    ):
        if not isinstance(condition, SagaOperation):
            condition = SagaOperation(condition)

        self.condition = condition
        self.saga = saga

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], IfThenAlternative], **kwargs) -> IfThenAlternative:
        """Build a new instance from a raw representation.

        :param raw: The raw representation.
        :param kwargs: Additional named arguments.
        :return: A new ``IfThenAlternative`` instance.
        """

        from ..saga import (
            Saga,
        )

        if isinstance(raw, cls):
            return raw
        current = raw | kwargs

        current["condition"] = SagaOperation.from_raw(current["condition"])
        current["saga"] = Saga.from_raw(current["saga"])

        return cls(**current)

    def validate(self) -> None:
        """Check if the alternative is valid.

        :return: This method does not return anything, but raises an exception if the alternative is not valid.
        """

        return self.saga.validate()

    @property
    def raw(self) -> dict[str, Any]:
        """Get the raw representation of the alternative.

        :return: A ``dict`` value.
        """

        return {
            "condition": self.condition.raw,
            "saga": self.saga.raw,
        }

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    def __iter__(self):
        yield from (
            self.condition,
            self.saga,
        )


class ElseThenAlternative:
    """Else Then Alternative class."""

    def __init__(self, saga: Saga):
        self.saga = saga

    @classmethod
    def from_raw(cls, raw: Union[dict[str, Any], ElseThenAlternative], **kwargs) -> ElseThenAlternative:
        """Build a new instance from a raw representation.

        :param raw: The raw representation.
        :param kwargs: Additional named arguments.
        :return: A new ``ElseThenAlternative`` instance.
        """

        from ..saga import (
            Saga,
        )

        if isinstance(raw, cls):
            return raw
        current = raw | kwargs

        current["saga"] = Saga.from_raw(current["saga"])

        return cls(**current)

    def validate(self) -> None:
        """Check if the alternative is valid.

        :return: This method does not return anything, but raises an exception if the alternative is not valid.
        """

        return self.saga.validate()

    @property
    def raw(self) -> dict[str, Any]:
        """Get the raw representation of the alternative.

        :return: A ``dict`` value.
        """

        return {
            "saga": self.saga.raw,
        }

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    def __iter__(self):
        yield from (self.saga,)



from .abc import (
    SagaStep,
)
from .conditional import (
    ConditionalSagaStep,
    ElseThenAlternative,
    IfThenAlternative,
)
from .local import (
    LocalSagaStep,
)
from .remote import (
    RemoteSagaStep,
)



from abc import (
    ABC,
    abstractmethod,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
    DatabaseOperationFactory,
)


class SagaExecutionDatabaseOperationFactory(DatabaseOperationFactory, ABC):
    """Saga Execution Database Operation Factory class."""

    @abstractmethod
    def build_store(self, uuid: UUID, **kwargs) -> DatabaseOperation:
        """Build the database operation to store a saga execution.

        :param uuid: The identifier of the saga execution.
        :param kwargs: The attributes of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        raise NotImplementedError

    @abstractmethod
    def build_load(self, uuid: UUID) -> DatabaseOperation:
        """Build the database operation to load a saga execution.

        :param uuid: The identifier of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        raise NotImplementedError

    @abstractmethod
    def build_delete(self, uuid: UUID) -> DatabaseOperation:
        """Build the database operation to delete a saga execution.

        :param uuid: The identifier of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        raise NotImplementedError



from .operations import (
    SagaOperation,
)
from .saga import (
    Saga,
)
from .steps import (
    ConditionalSagaStep,
    ElseThenAlternative,
    IfThenAlternative,
    LocalSagaStep,
    RemoteSagaStep,
    SagaStep,
)
from .types import (
    LocalCallback,
    RequestCallBack,
    ResponseCallBack,
)



from __future__ import (
    annotations,
)

import logging
import warnings
from functools import (
    reduce,
)
from operator import (
    or_,
)
from typing import (
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    Config,
    Inject,
    Injectable,
    NotProvidedException,
    PoolFactory,
    SetupMixin,
)
from minos.networks import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_USER_CONTEXT_VAR,
    BrokerClient,
    BrokerClientPool,
    BrokerMessage,
)

from .context import (
    SagaContext,
)
from .definitions import (
    Saga,
)
from .exceptions import (
    SagaFailedExecutionException,
    SagaPausedExecutionStepException,
)
from .executions import (
    DatabaseSagaExecutionRepository,
    SagaExecution,
    SagaExecutionRepository,
    SagaStatus,
)
from .messages import (
    SagaResponse,
)

logger = logging.getLogger(__name__)


@Injectable("saga_manager")
class SagaManager(SetupMixin):
    """Saga Manager implementation class.

    The purpose of this class is to manage the running process for new or paused``SagaExecution`` instances.
    """

    @Inject()
    def __init__(
        self,
        storage: SagaExecutionRepository,
        broker_pool: Optional[BrokerClientPool] = None,
        pool_factory: Optional[PoolFactory] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.storage = storage

        if broker_pool is None and pool_factory is not None:
            broker_pool = pool_factory.get_pool("broker")

        if broker_pool is None:
            raise NotProvidedException(f"A {BrokerClientPool!r} instance is required.")

        self.broker_pool = broker_pool

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> SagaManager:
        """Build an instance from config.

        :param args: Additional positional arguments.
        :param config: Config instance.
        :param kwargs: Additional named arguments.
        :return: A new ``SagaManager`` instance.
        """
        storage = DatabaseSagaExecutionRepository.from_config(config, **kwargs)
        return cls(storage=storage, **kwargs)

    async def _setup(self) -> None:
        await super()._setup()
        await self.storage.setup()

    async def _destroy(self) -> None:
        await self.storage.destroy()
        await super()._destroy()

    async def run(
        self,
        definition: Optional[Saga] = None,
        context: Optional[SagaContext] = None,
        *,
        response: Optional[SagaResponse] = None,
        user: Optional[UUID] = None,
        autocommit: bool = True,
        pause_on_disk: bool = False,
        raise_on_error: bool = True,
        return_execution: bool = True,
        **kwargs,
    ) -> Union[UUID, SagaExecution]:
        """Perform a run of a ``Saga``.

        The run can be a new one (if a name is provided) or continue execution a previous one (if a reply is provided).

        :param definition: Saga definition to be executed.
        :param context: Initial context to be used during the execution. (Only used for new executions)
        :param response: The reply that relaunches a saga execution.
        :param user: The user identifier to be injected on remote steps.
        :param autocommit: If ``True`` the transactions are committed/rejected automatically. Otherwise, the ``commit``
            or ``reject`` must to be called manually.
        :param pause_on_disk: If ``True`` the pauses until remote steps' responses are paused on disk (background,
            non-blocking the execution). Otherwise, the pauses are waited on memory (online, blocking the execution)
        :param raise_on_error: If ``True`` exceptions are raised on error. Otherwise, the execution is returned normally
            but with ``Errored`` status.
        :param return_execution: If ``True`` the ``SagaExecution`` instance is returned. Otherwise, only the
            identifier (``UUID``) is returned.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """

        if response is not None:
            return await self._load_and_run(
                response=response,
                autocommit=autocommit,
                pause_on_disk=pause_on_disk,
                raise_on_error=raise_on_error,
                return_execution=return_execution,
                **kwargs,
            )

        return await self._run_new(
            definition=definition,
            context=context,
            user=user,
            autocommit=autocommit,
            pause_on_disk=pause_on_disk,
            raise_on_error=raise_on_error,
            return_execution=return_execution,
            **kwargs,
        )

    async def _run_new(
        self, definition: Saga, context: Optional[SagaContext] = None, user: Optional[UUID] = None, **kwargs
    ) -> Union[UUID, SagaExecution]:
        if REQUEST_USER_CONTEXT_VAR.get() is not None:
            if user is not None:
                warnings.warn("The `user` Argument will be ignored in favor of the `user` ContextVar", RuntimeWarning)
            user = REQUEST_USER_CONTEXT_VAR.get()

        execution = SagaExecution.from_definition(definition, context=context, user=user)
        return await self._run(execution, **kwargs)

    async def _load_and_run(self, response: SagaResponse, **kwargs) -> Union[UUID, SagaExecution]:
        execution = await self.storage.load(response.uuid)
        return await self._run(execution, response=response, **kwargs)

    async def _run(
        self,
        execution: SagaExecution,
        pause_on_disk: bool = False,
        raise_on_error: bool = True,
        return_execution: bool = True,
        **kwargs,
    ) -> Union[UUID, SagaExecution]:
        try:
            if pause_on_disk:
                await self._run_with_pause_on_disk(execution, **kwargs)
            else:
                await self._run_with_pause_on_memory(execution, **kwargs)
        except SagaFailedExecutionException as exc:
            await self.storage.store(execution)
            if raise_on_error:
                raise exc
            logger.exception(f"The execution identified by {execution.uuid!s} failed")
        finally:
            if (headers := REQUEST_HEADERS_CONTEXT_VAR.get()) is not None:
                related_services = reduce(or_, (s.related_services for s in execution.executed_steps), set())
                if execution.paused_step is not None:
                    related_services.update(execution.paused_step.related_services)

                if raw_related_services := headers.get("related_services"):
                    related_services.update(raw_related_services.split(","))

                headers["related_services"] = ",".join(related_services)

        if execution.status == SagaStatus.Finished:
            await self.storage.delete(execution)

        if return_execution:
            return execution

        return execution.uuid

    async def _run_with_pause_on_disk(self, execution: SagaExecution, autocommit: bool = True, **kwargs) -> None:
        try:
            await execution.execute(autocommit=False, **kwargs)
            if autocommit:
                await execution.commit(**kwargs)
        except SagaPausedExecutionStepException:
            await self.storage.store(execution)
        except SagaFailedExecutionException as exc:
            if autocommit:
                await execution.reject(**kwargs)
            raise exc

    async def _run_with_pause_on_memory(
        self, execution: SagaExecution, response: Optional[SagaResponse] = None, autocommit: bool = True, **kwargs
    ) -> None:

        try:
            # noinspection PyUnresolvedReferences
            async with self.broker_pool.acquire() as broker:
                while execution.status in (SagaStatus.Created, SagaStatus.Paused):
                    try:
                        await execution.execute(response=response, autocommit=False, **kwargs)
                    except SagaPausedExecutionStepException:
                        response = await self._get_response(broker, execution, **kwargs)
                    await self.storage.store(execution)
            if autocommit:
                await execution.commit(**kwargs)
        except SagaFailedExecutionException as exc:
            if autocommit:
                await execution.reject(**kwargs)
            raise exc

    @staticmethod
    async def _get_response(broker: BrokerClient, execution: SagaExecution, **kwargs) -> SagaResponse:
        message: Optional[BrokerMessage] = None
        while message is None or "saga" not in message.headers or UUID(message.headers["saga"]) != execution.uuid:
            try:
                message = await broker.receive(**kwargs)
            except Exception as exc:
                execution.status = SagaStatus.Errored
                raise SagaFailedExecutionException(exc)
        return SagaResponse.from_message(message)



from __future__ import (
    annotations,
)

from typing import (
    Any,
    Iterable,
    Optional,
    Union,
)

from minos.common import (
    classname,
)

from ...context import (
    SagaContext,
)
from ...exceptions import (
    EmptySagaStepException,
    MultipleOnErrorException,
    MultipleOnExecuteException,
    MultipleOnFailureException,
    MultipleOnSuccessException,
    UndefinedOnExecuteException,
)
from ..operations import (
    SagaOperation,
)
from ..types import (
    RequestCallBack,
    ResponseCallBack,
)
from .abc import (
    SagaStep,
)


class RemoteSagaStep(SagaStep):
    """Remote Saga Step class."""

    def __init__(
        self,
        on_execute: Optional[Union[RequestCallBack, SagaOperation[RequestCallBack]]] = None,
        on_success: Optional[Union[ResponseCallBack, SagaOperation[ResponseCallBack]]] = None,
        on_error: Optional[Union[ResponseCallBack, SagaOperation[ResponseCallBack]]] = None,
        on_failure: Optional[Union[RequestCallBack, SagaOperation[RequestCallBack]]] = None,
        **kwargs,
    ):

        if on_execute is not None and not isinstance(on_execute, SagaOperation):
            on_execute = SagaOperation(on_execute)
        if on_failure is not None and not isinstance(on_failure, SagaOperation):
            on_failure = SagaOperation(on_failure)
        if on_success is not None and not isinstance(on_success, SagaOperation):
            on_success = SagaOperation(on_success)
        if on_error is not None and not isinstance(on_error, SagaOperation):
            on_error = SagaOperation(on_error)

        self.on_execute_operation = on_execute
        self.on_failure_operation = on_failure
        self.on_success_operation = on_success
        self.on_error_operation = on_error

        super().__init__(**kwargs)

    @classmethod
    def _from_raw(cls, raw: dict[str, Any]) -> RemoteSagaStep:
        raw["on_execute"] = SagaOperation.from_raw(raw["on_execute"])
        raw["on_failure"] = SagaOperation.from_raw(raw["on_failure"])
        raw["on_success"] = SagaOperation.from_raw(raw["on_success"])
        raw["on_error"] = SagaOperation.from_raw(raw["on_error"])

        return cls(**raw)

    def on_execute(
        self, callback: RequestCallBack, parameters: Optional[SagaContext] = None, **kwargs
    ) -> RemoteSagaStep:
        """On execute method.

        :param callback: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_execute_operation is not None:
            raise MultipleOnExecuteException()

        self.on_execute_operation = SagaOperation(callback, parameters, **kwargs)

        return self

    def on_failure(
        self, callback: RequestCallBack, parameters: Optional[SagaContext] = None, **kwargs
    ) -> RemoteSagaStep:
        """On failure method.

        :param callback: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_failure_operation is not None:
            raise MultipleOnFailureException()

        self.on_failure_operation = SagaOperation(callback, parameters, **kwargs)

        return self

    def on_success(
        self, callback: ResponseCallBack, parameters: Optional[SagaContext] = None, **kwargs
    ) -> RemoteSagaStep:
        """On success method.

        :param callback: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_success_operation is not None:
            raise MultipleOnSuccessException()

        self.on_success_operation = SagaOperation(callback, parameters, **kwargs)

        return self

    def on_error(
        self, callback: ResponseCallBack, parameters: Optional[SagaContext] = None, **kwargs
    ) -> RemoteSagaStep:
        """On error method.

        :param callback: The callback function to be called.
        :param parameters: A mapping of named parameters to be passed to the callback.
        :param kwargs: A set of named arguments to be passed to the callback. ``parameters`` has priority if it is not
            ``None``.
        :return: A ``self`` reference.
        """
        if self.on_error_operation is not None:
            raise MultipleOnErrorException()

        self.on_error_operation = SagaOperation(callback, parameters, **kwargs)

        return self

    def validate(self) -> None:
        """Check if the step is valid.

        :return: This method does not return anything, but raises an exception if the step is not valid.
        """
        if (
            self.on_execute_operation is None
            and self.on_failure_operation is None
            and self.on_success_operation is None
            and self.on_error_operation is None
        ):
            raise EmptySagaStepException()

        if self.on_execute_operation is None:
            raise UndefinedOnExecuteException()

    @property
    def raw(self) -> dict[str, Any]:
        """Generate a raw representation of the instance.

        :return: A ``dict`` instance.
        """
        return {
            "cls": classname(type(self)),
            "on_execute": None if self.on_execute_operation is None else self.on_execute_operation.raw,
            "on_failure": None if self.on_failure_operation is None else self.on_failure_operation.raw,
            "on_success": None if self.on_success_operation is None else self.on_success_operation.raw,
            "on_error": None if self.on_error_operation is None else self.on_error_operation.raw,
        }

    def __iter__(self) -> Iterable:
        yield from (
            self.on_execute_operation,
            self.on_failure_operation,
            self.on_success_operation,
            self.on_error_operation,
        )
